<?php
/**
 * Handle php errors
 * For each error show backtrace if it's possible
 * 
 */
class ErrorHandler {
	
	const LEVEL_RUNTIME  = 'Runtime';
	const LEVEL_SHUTDOWN = 'Shutdown';
	public static $show = false;
	public static $registered = false;
	
	public static function register() {
		if(self::$registered) {
			return;
		}
		register_shutdown_function(array(__CLASS__, 'handle_shutdown'));
		set_error_handler(array(__CLASS__, 'handle_runtime'));
		self::$show = (bool)ini_get("display_errors");
	}
	public static function unregister() {
		if(!self::$registered) {
			return;
		}
		restore_error_handler();
	}
	/**
	 * @see register_shutdown_function()
	 */
	public static function handle_shutdown() {
		$last_error = error_get_last();
		if(empty($last_error)) {
			return;
		}
		switch($last_error['type']) {
			//do not handle errors that we can catch with set_error_handler @see self::handle_runtime
			case E_NOTICE:
			case E_DEPRECATED:
			case E_STRICT:
			case E_WARNING:
			case E_CORE_WARNING:
			case E_USER_NOTICE:
			case E_USER_WARNING:
			case E_USER_ERROR:
			case E_RECOVERABLE_ERROR:
				break;
			default:
				self::log_error(
					self::LEVEL_SHUTDOWN,
					$last_error['type'],
					$last_error['message'],
					$last_error['file'],
					$last_error['line']
				);
		}
		
	}
	/**
	 * @see set_error_handler()
	 */
	public static function handle_runtime($type, $message, $file, $line) {
		// error_reporting value will be 0
		// if the statement that caused the error was prepended by the @ error-control operator.
	    if (0 == error_reporting()) {
	        return false;
    	}
	    switch ($type) {
			case E_WARNING:
			case E_USER_WARNING:
			case E_USER_ERROR:
			case E_RECOVERABLE_ERROR:
				self::log_error(
					self::LEVEL_RUNTIME,
					$type,
					$message,
					$file,
					$line
				);
				switch($type) {
					case E_WARNING:
					case E_USER_WARNING:
						return true; // continue normal execution
					default:
						if(self::$show) {
							die;
						}
				}
	    		return true;
		}
	    return false;
	    //the normal error handler continues.
	}
	/**
	 * @param string $level @see ETB_ErrorHandler::LEVEL_*
	 * @param int $code @see E_*
	 * @param string $message
	 * @param string $file path
	 * @param int $line
	 */
	private static function log_error($level, $type, $message, $file, $line) {
		$error_name = self::get_error_name($type);
		$source = isset($_SERVER['REQUEST_URI']) ? $_SERVER['REQUEST_URI'] : implode(' ', $_SERVER['argv']);
		$message =
			 "PHP $level $error_name: $message"
			.PHP_EOL."File: $file($line)"
			.PHP_EOL."Request: {$source}"
			.(isset($_SERVER['HTTP_REFERER']) ? PHP_EOL."Referer: {$_SERVER['HTTP_REFERER']}" : '')
		;
		if($level == self::LEVEL_RUNTIME) {
			$message .= PHP_EOL.'Backtrace: '.PHP_EOL.self::get_backtrace();
		}
		error_log($message);
		if(self::$show) {
			echo $message,PHP_EOL,PHP_EOL;
		}
	}

	/**
	 * @param $error_code @see E_*
	 */
	private static function get_error_name($error_code) {
		switch ($error_code) {
			case E_ERROR: case E_CORE_ERROR: case E_COMPILE_ERROR: case E_USER_ERROR:
				return "Fatal Error";
			case E_RECOVERABLE_ERROR:
				return "Catchable Error";
			case E_WARNING: case E_CORE_WARNING: case E_COMPILE_WARNING: case E_USER_WARNING:
				return "Warning";
			case E_PARSE:
				return "Parse Error";
			case E_NOTICE: case E_USER_NOTICE:
				return "Notice";
			case E_STRICT:
				return "Strict Notice";
		}
		return 'Unknown Error';
	}
	/**
	 * @return string
	 * @example
	 * 		#0  /tmp/include.php(5): eval()
	 * 		#1  /tmp/include.php(17): a()
	 * 		#2  /tmp/test.php(3): include(/tmp/include.php)
	 * 		#3  {main}
	 */
	public static function get_backtrace() {
		$trace = debug_backtrace();
	
		$length = count($trace);
		$result = '';
		for ($i=3;$i<$length;$i++) { // $i=3 skip error handler functions
			$args = array();
			for ($j=0;$j<count($trace[$i]['args']);$j++) {
				$args[]=(is_object($trace[$i]['args'][$j]))?get_class($trace[$i]['args'][$j]):substr((string)$trace[$i]['args'][$j],0,100);
			}
			$args=implode(',',$args);
			$args=($args)?'('.$args.')':'';
			$result.= '#'.($i-2).' '.$trace[$i]['file'].'('.$trace[$i]['line'].'):'.$trace[$i]['function'].$args."\n";
		}
		if ($result) {
			$result.= '#'.($i-2)." {main}";
		}
		return $result;
	}
}